@@ -123,45 +123,123 @@ get_creator (NMHalManager *self, const char *udi)
 
 /* Common helpers for built-in device creators */
 
+static char *
+hal_get_subsystem (LibHalContext *ctx, const char *udi)
+{
+	char *subsys;
+
+	subsys = libhal_device_get_property_string (ctx, udi, "info.subsystem", NULL);
+	if (!subsys) {
+		/* info.bus is deprecated */
+		subsys = libhal_device_get_property_string (ctx, udi, "info.bus", NULL);
+	}
+	return subsys;
+}
+
 static char *
 nm_get_device_driver_name (LibHalContext *ctx, const char *origdev_udi)
 {
-	char *driver_name = NULL;
+	char *driver_name = NULL, *subsystem, *drv, *od_parent = NULL;
 
-	if (origdev_udi && libhal_device_property_exists (ctx, origdev_udi, "info.linux.driver", NULL)) {
-		char *drv;
+	if (!origdev_udi)
+		return NULL;
 
-		drv = libhal_device_get_property_string (ctx, origdev_udi, "info.linux.driver", NULL);
-		if (drv) {
-			driver_name = g_strdup (drv);
-			libhal_free_string (drv);
+	/* s390 driver name is on the grandparent of the net device */
+	subsystem = hal_get_subsystem (ctx, origdev_udi);
+	if (subsystem) {
+		if (!strcmp (subsystem, "ibmebus")) {
+			od_parent = libhal_device_get_property_string (ctx, origdev_udi, "info.parent", NULL);
+			origdev_udi = (const char *) od_parent;
+		} else if (!strcmp (subsystem, "ssb")) {
+			od_parent = libhal_device_get_property_string (ctx, origdev_udi, "info.parent", NULL);
+			origdev_udi = (const char *) od_parent;
 		}
 	}
+
+	drv = libhal_device_get_property_string (ctx, origdev_udi, "info.linux.driver", NULL);
+	if (drv)
+		driver_name = g_strdup (drv);
+
+	libhal_free_string (drv);
+	libhal_free_string (od_parent);
+	libhal_free_string (subsystem);
 	return driver_name;
 }
 
-/* Returns the parent if the device is a Sony Ericsson 'mbm'-style device */
-static char *
-is_mbm (LibHalContext *ctx, const char *udi)
+static gboolean
+get_vid_pid (LibHalContext *ctx,
+             const char *udi,
+             guint16 *out_vid,
+             guint16 *out_pid,
+             char **out_parent)
 {
-	guint32 vendor_id = 0, product_id = 0;
+	int tmp;
 	char *parent;
+	DBusError error;
+	gboolean success = FALSE;
+
+	g_return_val_if_fail (ctx != NULL, FALSE);
+	g_return_val_if_fail (udi != NULL, FALSE);
 
 	parent = libhal_device_get_property_string (ctx, udi, "info.parent", NULL);
 	if (!parent)
-		return NULL;
+		return FALSE;
 
-	vendor_id = libhal_device_get_property_int (ctx, parent, "usb.vendor_id", NULL);
-	product_id = libhal_device_get_property_int (ctx, parent, "usb.product_id", NULL);
+	dbus_error_init (&error);
+	tmp = libhal_device_get_property_int (ctx, parent, "usb.vendor_id", &error);
+	if (dbus_error_is_set (&error)) {
+		dbus_error_free (&error);
+		goto out;
+	}
+	if (tmp < 0 || tmp > G_MAXUINT16)
+		goto out;
+	if (out_vid)
+		*out_vid = (guint16) tmp;
 
-	if (   (vendor_id == 0x0bdb && product_id == 0x1900)  /* SE F3507g */
-	    || (vendor_id == 0x0bdb && product_id == 0x1902)  /* SE F3507g */
-	    || (vendor_id == 0x0fce && product_id == 0xd0cf)  /* SE MD300 */
-	    || (vendor_id == 0x413c && product_id == 0x8147)) /* Dell 5530 HSDPA */
-		return parent;
+	dbus_error_init (&error);
+	tmp = libhal_device_get_property_int (ctx, parent, "usb.product_id", &error);
+	if (dbus_error_is_set (&error)) {
+		dbus_error_free (&error);
+		goto out;
+	}
+	if (tmp < 0 || tmp > G_MAXUINT16)
+		goto out;
+	if (out_pid)
+		*out_pid = (guint16) tmp;
 
+	if (out_parent)
+		*out_parent = g_strdup (parent);
+	success = TRUE;
+
+out:
 	libhal_free_string (parent);
-	return NULL;
+	return success;
+}
+
+/* Returns the parent if the device is a Sony Ericsson 'mbm'-style device */
+static gboolean
+is_mbm (LibHalContext *ctx, const char *udi, char **out_parent)
+{
+	guint16 vendor_id = 0, product_id = 0;
+	char *parent = NULL;
+
+	if (!get_vid_pid (ctx, udi, &vendor_id, &product_id, &parent))
+		return FALSE;
+
+	if (   (vendor_id == 0x0bdb && product_id == 0x1900)    /* SE F3507g */
+	    || (vendor_id == 0x0bdb && product_id == 0x1902)    /* SE F3507g */
+	    || (vendor_id == 0x0bdb && product_id == 0x1904)    /* SE F3607gw */
+	    || (vendor_id == 0x0bdb && product_id == 0x1906)    /* SE F3307 */
+	    || (vendor_id == 0x0fce && product_id == 0xd0cf)    /* SE MD300 */
+	    || (vendor_id == 0x413c && product_id == 0x8147)    /* Dell 5530 HSDPA */
+	    || (vendor_id == 0x0930 && product_id == 0x130b)) { /* Toshiba */
+		if (out_parent)
+			*out_parent = parent;
+		return TRUE;
+	}
+
+	g_free (parent);
+	return FALSE;
 }
 
 /* Wired device creator */
@@ -194,8 +272,7 @@ wired_device_creator (NMHalManager *self,
 {
 	NMHalManagerPrivate *priv = NM_HAL_MANAGER_GET_PRIVATE (self);
 	GObject *device = NULL;
-	char *iface, *driver, *parent;
-	gboolean mbm = FALSE;
+	char *iface, *driver;
 
 	iface = libhal_device_get_property_string (priv->hal_ctx, udi, "net.interface", NULL);
 	if (!iface) {
@@ -208,13 +285,7 @@ wired_device_creator (NMHalManager *self,
 	/* Special handling of Ericsson F3507g 'mbm' devices; ignore the
 	 * cdc-ether device that it provides since we don't use it yet.
 	 */
-	if (driver && !strcmp (driver, "cdc_ether")) {
-		parent = is_mbm (priv->hal_ctx, udi);
-		mbm = !!parent;
-		libhal_free_string (parent);
-	}
-
-	if (!mbm)
+	if (!is_mbm (priv->hal_ctx, udi, NULL))
 		device = (GObject *) nm_device_ethernet_new (udi, iface, driver, managed);
 
 	libhal_free_string (iface);
@@ -291,19 +362,35 @@ is_modem_device (NMHalManager *self, const char *udi)
 	return is_modem;
 }
 
+static char *
+hal_get_originating_device (LibHalContext *ctx, const char *udi, const char *prefix)
+{
+	char *od_key = g_strdup_printf ("%s.originating_device", prefix);
+	char *pd_key = g_strdup_printf ("%s.physical_device", prefix);
+	char *od;
+
+	od = libhal_device_get_property_string (ctx, udi, od_key, NULL);
+	if (!od) {
+		/* physical_device is deprecated */
+		od = libhal_device_get_property_string (ctx, udi, pd_key, NULL);
+	}
+
+	g_free (od_key);
+	g_free (pd_key);
+	return od;
+}
+
 static char *
 get_hso_netdev (LibHalContext *ctx, const char *udi)
 {
-	char *serial_od = NULL, *serial_od_parent = NULL, *netdev = NULL, *bus;
+	char *serial_od = NULL, *serial_od_parent = NULL, *netdev = NULL, *subsys;
 	char **netdevs = NULL;
 	int num, i;
 
 	/* Get the serial interface's originating device UDI, used to find the
 	 * originating device's netdev.
 	 */
-	serial_od = libhal_device_get_property_string (ctx, udi, "serial.originating_device", NULL);
-	if (!serial_od)
-		serial_od = libhal_device_get_property_string (ctx, udi, "serial.physical_device", NULL);
+	serial_od = hal_get_originating_device (ctx, udi, "serial");
 	if (!serial_od)
 		goto out;
 
@@ -312,21 +399,19 @@ get_hso_netdev (LibHalContext *ctx, const char *udi)
 		goto out;
 
 	/* Check to ensure we've got the actual "USB Device" */
-	bus = libhal_device_get_property_string (ctx, serial_od_parent, "info.bus", NULL);
-	if (!bus || strcmp (bus, "usb_device")) {
-		libhal_free_string (bus);
+	subsys = hal_get_subsystem (ctx, serial_od_parent);
+	if (!subsys || strcmp (subsys, "usb_device")) {
+		libhal_free_string (subsys);
 		goto out;
 	}
-	libhal_free_string (bus);
+	libhal_free_string (subsys);
 
 	/* Look for the originating device's netdev */
 	netdevs = libhal_find_device_by_capability (ctx, "net", &num, NULL);
 	for (i = 0; netdevs && !netdev && (i < num); i++) {
 		char *net_od = NULL, *net_od_parent = NULL, *tmp;
 
-		net_od = libhal_device_get_property_string (ctx, netdevs[i], "net.originating_device", NULL);
-		if (!net_od)
-			net_od = libhal_device_get_property_string (ctx, netdevs[i], "net.physical_device", NULL);
+		net_od = hal_get_originating_device (ctx, netdevs[i], "net");
 		if (!net_od)
 			goto next;
 
@@ -335,12 +420,12 @@ get_hso_netdev (LibHalContext *ctx, const char *udi)
 			goto next;
 
 		/* Check to ensure we've got the actual "USB Device" */
-		bus = libhal_device_get_property_string (ctx, net_od_parent, "info.bus", NULL);
-		if (!bus || strcmp (bus, "usb_device")) {
-			libhal_free_string (bus);
+		subsys = hal_get_subsystem (ctx, net_od_parent);
+		if (!subsys || strcmp (subsys, "usb_device")) {
+			libhal_free_string (subsys);
 			goto next;
 		}
-		libhal_free_string (bus);
+		libhal_free_string (subsys);
 
 		if (!strcmp (net_od_parent, serial_od_parent)) {
 			/* We found it */
@@ -363,16 +448,20 @@ out:
 	return netdev;
 }
 
-#define PROP_GSM   "ID_NM_MODEM_GSM"
-#define PROP_CDMA  "ID_NM_MODEM_IS707_A"
-#define PROP_EVDO1 "ID_NM_MODEM_IS856"
-#define PROP_EVDOA "ID_NM_MODEM_IS856_A"
+#define PROP_GSM      "ID_NM_MODEM_GSM"
+#define PROP_CDMA     "ID_NM_MODEM_IS707_A"
+#define PROP_EVDO1    "ID_NM_MODEM_IS856"
+#define PROP_EVDOA    "ID_NM_MODEM_IS856_A"
+#define PROP_ZTE_MAIN "ID_NM_ZTE_PORT_TYPE_MODEM"
+#define PROP_ZTE_SET  "ID_NM_ZTE_PORT_TYPE_SET"
 
 #if HAVE_LIBUDEV
 
 typedef struct {
 	gboolean gsm;
 	gboolean cdma;
+	gboolean zte_main;
+	gboolean zte_set;
 } UdevIterData;
 
 #if UDEV_VERSION >= 129
@@ -405,6 +494,13 @@ static int udev_device_prop_iter(struct udev_device *udev_device,
 	if (!strcmp (key, PROP_EVDOA) && !strcmp (value, "1"))
 		types->cdma = TRUE;
 
+	if (types->cdma || types->gsm) {
+	    if (!strcmp (key, PROP_ZTE_MAIN) && !strcmp (value, "1"))
+			types->zte_main = TRUE;
+	    if (!strcmp (key, PROP_ZTE_SET) && !strcmp (value, "1"))
+			types->zte_set = TRUE;
+	}
+
 	/* Return 0 to continue looking */
 	return types->gsm && types->cdma;
 }
@@ -413,7 +509,9 @@ static int udev_device_prop_iter(struct udev_device *udev_device,
 static gboolean
 libudev_get_modem_capabilities (const char *sysfs_path,
                                 gboolean *gsm,
-                                gboolean *cdma)
+                                gboolean *cdma,
+                                gboolean *zte_main,
+                                gboolean *zte_set)
 {
 	struct udev *udev;
 	struct udev_device *device;
@@ -423,6 +521,10 @@ libudev_get_modem_capabilities (const char *sysfs_path,
 	g_return_val_if_fail (*gsm == FALSE, FALSE);
 	g_return_val_if_fail (cdma != NULL, FALSE);
 	g_return_val_if_fail (*cdma == FALSE, FALSE);
+	g_return_val_if_fail (zte_main != NULL, FALSE);
+	g_return_val_if_fail (*zte_main == FALSE, FALSE);
+	g_return_val_if_fail (zte_set != NULL, FALSE);
+	g_return_val_if_fail (*zte_set == FALSE, FALSE);
 
 	udev = udev_new ();
 	if (!udev)
@@ -450,6 +552,8 @@ libudev_get_modem_capabilities (const char *sysfs_path,
 		const char *cdma_val = get_udev_property (device, PROP_CDMA);
 		const char *evdo1_val = get_udev_property (device, PROP_EVDO1);
 		const char *evdoa_val = get_udev_property (device, PROP_EVDOA);
+		const char *zte_main_val = get_udev_property (device, PROP_ZTE_MAIN);
+		const char *zte_set_val = get_udev_property (device, PROP_ZTE_SET);
 
 		if (gsm_val && !strcmp (gsm_val, "1"))
 			*gsm = TRUE;
@@ -459,6 +563,13 @@ libudev_get_modem_capabilities (const char *sysfs_path,
 			*cdma = TRUE;
 		if (evdoa_val && !strcmp (evdoa_val, "1"))
 			*cdma = TRUE;
+
+		if (*gsm || *cdma) {
+		    if (zte_main_val && !strcmp (zte_main_val, "1"))
+				*zte_main = TRUE;
+		    if (zte_set_val && !strcmp (zte_set_val, "1"))
+				*zte_set = TRUE;
+		}
 	}
 #else
 	{
@@ -467,6 +578,10 @@ libudev_get_modem_capabilities (const char *sysfs_path,
 		udev_device_get_properties (device, udev_device_prop_iter, &iterdata);
 		*gsm = iterdata.gsm;
 		*cdma = iterdata.cdma;
+		if (*gsm || *cdma) {
+			*zte_main = iterdata.zte_main;
+			*zte_set = iterdata.zte_set;
+		}
 	}
 #endif
 
@@ -478,7 +593,9 @@ libudev_get_modem_capabilities (const char *sysfs_path,
 static gboolean
 udevadm_get_modem_capabilities (const char *sysfs_path,
                                 gboolean *gsm,
-                                gboolean *cdma)
+                                gboolean *cdma,
+                                gboolean *zte_main,
+                                gboolean *zte_set)
 {
 	char *udevadm_argv[] = { "/sbin/udevadm", "info", "--query=env", NULL, NULL };
 	char *syspath_arg = NULL;
@@ -486,13 +603,17 @@ udevadm_get_modem_capabilities (const char *sysfs_path,
 	int exitcode;
 	GError *error = NULL;
 	char **lines = NULL, **iter;
-	gboolean success = FALSE;
+	gboolean success = FALSE, is_zte_main = FALSE, is_zte_set = FALSE;
 
 	g_return_val_if_fail (sysfs_path != NULL, FALSE);
 	g_return_val_if_fail (gsm != NULL, FALSE);
 	g_return_val_if_fail (*gsm == FALSE, FALSE);
 	g_return_val_if_fail (cdma != NULL, FALSE);
 	g_return_val_if_fail (*cdma == FALSE, FALSE);
+	g_return_val_if_fail (zte_main != NULL, FALSE);
+	g_return_val_if_fail (*zte_main == FALSE, FALSE);
+	g_return_val_if_fail (zte_set != NULL, FALSE);
+	g_return_val_if_fail (*zte_set == FALSE, FALSE);
 
 	udevadm_argv[3] = syspath_arg = g_strdup_printf ("--path=%s", sysfs_path);
 	if (g_spawn_sync ("/", udevadm_argv, NULL, 0, NULL, NULL,
@@ -517,14 +638,21 @@ udevadm_get_modem_capabilities (const char *sysfs_path,
 	for (iter = lines; *iter; iter++) {
 		if (!strcmp (*iter, PROP_GSM "=1")) {
 			*gsm = TRUE;
-			break;
 		} else if (   !strcmp (*iter, PROP_CDMA "=1")
 		           || !strcmp (*iter, PROP_EVDO1 "=1")
 		           || !strcmp (*iter, PROP_EVDOA "=1")) {
 			*cdma = TRUE;
-			break;
-		}
+		} else if (!strcmp (*iter, PROP_ZTE_MAIN "=1"))
+			is_zte_main = TRUE;
+		else if (!strcmp (*iter, PROP_ZTE_SET "=1"))
+			is_zte_set = TRUE;
 	}
+
+	if (*gsm || *cdma) {
+		*zte_main = is_zte_main;
+		*zte_set = is_zte_set;
+	}
+
 	success = TRUE;
 
 error:
@@ -823,13 +951,12 @@ new_modem_device (const char *udi,
 	 * ModemManager more correctly in HEAD.
 	 */
 	if (gsm && !strcmp (driver, "cdc_acm")) {
-		char *parent;
+		char *parent = NULL;
 		guint32 usb_interface;
 
-		parent = is_mbm (ctx, udi);
-		if (parent) {
+		if (is_mbm (ctx, udi, &parent)) {
 			usb_interface = libhal_device_get_property_int (ctx, parent, "usb.interface.number", NULL);
-			libhal_free_string (parent);
+			g_free (parent);
 			if (usb_interface != 1)
 				return NULL;
 		}
@@ -851,22 +978,16 @@ new_modem_device (const char *udi,
 static char *
 nm_get_modem_device_driver_name (LibHalContext *ctx, const char *udi)
 {
-	char *driver_name = NULL;
-	char *origdev_udi;
+	char *driver_name = NULL, *origdev_udi, *driver;
 
-	origdev_udi = libhal_device_get_property_string (ctx, udi, "serial.originating_device", NULL);
-	/* Older HAL uses "physical_device" */
+	origdev_udi = hal_get_originating_device (ctx, udi, "serial");
 	if (!origdev_udi)
-		origdev_udi = libhal_device_get_property_string (ctx, udi, "serial.physical_device", NULL);
+		return NULL;
 
-	if (origdev_udi && libhal_device_property_exists (ctx, origdev_udi, "info.linux.driver", NULL)) {
-		char *drv;
-
-		drv = libhal_device_get_property_string (ctx, origdev_udi, "info.linux.driver", NULL);
-		if (drv) {
-			driver_name = g_strdup (drv);
-			libhal_free_string (drv);
-		}
+	driver = libhal_device_get_property_string (ctx, origdev_udi, "info.linux.driver", NULL);
+	if (driver) {
+		driver_name = g_strdup (driver);
+		libhal_free_string (driver);
 	}
 	libhal_free_string (origdev_udi);
 	return driver_name;
@@ -890,7 +1011,10 @@ modem_device_creator (NMHalManager *self,
 	gboolean hal_cdma = FALSE;
 	gboolean later = FALSE;
 	gboolean udev_success = FALSE;
+	gboolean zte_main = FALSE;
+	gboolean zte_set = FALSE;
 	DeferredModem *deferred = NULL;
+	guint16 vid = 0;
 
 	serial_device = libhal_device_get_property_string (priv->hal_ctx, udi, "serial.device", NULL);
 	/* For serial devices, 'origdev_udi' will be the actual USB or platform device,
@@ -921,9 +1045,9 @@ modem_device_creator (NMHalManager *self,
 	}
 
 #if HAVE_LIBUDEV
-	udev_success = libudev_get_modem_capabilities (sysfs_path, &udev_gsm, &udev_cdma);
+	udev_success = libudev_get_modem_capabilities (sysfs_path, &udev_gsm, &udev_cdma, &zte_main, &zte_set);
 #else
-	udev_success = udevadm_get_modem_capabilities (sysfs_path, &udev_gsm, &udev_cdma);
+	udev_success = udevadm_get_modem_capabilities (sysfs_path, &udev_gsm, &udev_cdma, &zte_main, &zte_set);
 #endif
 	libhal_free_string (sysfs_path);
 
@@ -937,6 +1061,17 @@ modem_device_creator (NMHalManager *self,
 		udev_cdma = hal_cdma;
 	}
 
+	/* ZTE modem ports get tagged by udev rules, so if the modem was known to
+	 * the rules (in which case ID_NM_ZTE_PORT_TYPE_SET will be set), and the
+	 * port isn't a "main" port, we should ignore it.
+	 */
+	if (get_vid_pid (priv->hal_ctx, udi, &vid, NULL, NULL) && (vid == 0x19d2)) {
+		if (zte_set && !zte_main) {
+			udev_gsm = FALSE;
+			udev_cdma = FALSE;
+		}
+	}
+
 	/* If it's not known to either udev or HAL as a modem, nothing to do */
 	if (!udev_gsm && !udev_cdma && !hal_gsm && !hal_cdma) {
 		nm_info ("(%s): ignoring due to lack of mobile broadband capabilties", ttyname);
@@ -989,7 +1124,7 @@ modem_device_creator (NMHalManager *self,
 
 		if (priv->deferred_modem_id)
 			g_source_remove (priv->deferred_modem_id);
-		priv->deferred_modem_id = g_timeout_add_seconds (4, deferred_modem_timeout, self);
+		priv->deferred_modem_id = g_timeout_add_seconds (8, deferred_modem_timeout, self);
 
 		nm_info ("(%s): deferring until all ports found", ttyname);
 	}
@@ -1039,7 +1174,7 @@ static void
 emit_udi_added (NMHalManager *self, const char *udi, DeviceCreator *creator)
 {
 	NMHalManagerPrivate *priv = NM_HAL_MANAGER_GET_PRIVATE (self);
-	char *od = NULL, *tmp, *parent, *bus = NULL;
+	char *od = NULL, *tmp, *parent, *subsys = NULL;
 
 	g_return_if_fail (self != NULL);
 	g_return_if_fail (udi != NULL);
@@ -1056,8 +1191,8 @@ emit_udi_added (NMHalManager *self, const char *udi, DeviceCreator *creator)
 	 */
 	parent = libhal_device_get_property_string (priv->hal_ctx, udi, "info.parent", NULL);
 	if (parent)
-		bus = libhal_device_get_property_string (priv->hal_ctx, parent, "info.bus", NULL);
-	if (bus && !strcmp (bus, "usb")) {
+		subsys = hal_get_subsystem (priv->hal_ctx, parent);
+	if (subsys && !strcmp (subsys, "usb")) {
 		char *usb_intf_udi;
 
 		usb_intf_udi = libhal_device_get_property_string (priv->hal_ctx, udi, "info.parent", NULL);
@@ -1066,7 +1201,7 @@ emit_udi_added (NMHalManager *self, const char *udi, DeviceCreator *creator)
 
 			/* Ensure the grandparent really is the "USB Device" */
 			if (od) {
-				tmp = libhal_device_get_property_string (priv->hal_ctx, od, "info.bus", NULL);
+				tmp = hal_get_subsystem (priv->hal_ctx, od);
 				if (!tmp || strcmp (tmp, "usb_device")) {
 					libhal_free_string (od);
 					od = NULL;
@@ -1077,24 +1212,15 @@ emit_udi_added (NMHalManager *self, const char *udi, DeviceCreator *creator)
 			libhal_free_string (usb_intf_udi);
 		}
 	}
-	libhal_free_string (bus);
+	libhal_free_string (subsys);
 	libhal_free_string (parent);
 
-	/* For non-USB devices, and ss a fallback, just use the originating device
-	 * of the tty; though this might result in more than one modem being detected by NM.
+	/* For non-USB devices, and as a fallback, just use the originating device
+	 * of the tty; though this might result in more than one modem being
+	 * detected by NM.
 	 */
-	if (!od) {
-		tmp = g_strdup_printf ("%s.originating_device", creator->category);
-		od = libhal_device_get_property_string (priv->hal_ctx, udi, tmp, NULL);
-		g_free (tmp);
-	}
-
-	if (!od) {
-		/* Older HAL uses 'physical_device' */
-		tmp = g_strdup_printf ("%s.physical_device", creator->category);
-		od = libhal_device_get_property_string (priv->hal_ctx, udi, tmp, NULL);
-		g_free (tmp);
-	}
+	if (!od)
+		od = hal_get_originating_device (priv->hal_ctx, udi, creator->category);
 
 	g_signal_emit (self, signals[UDI_ADDED], 0,
 	               udi,
